const std = @import("std");
const mem = std.mem;
const utils = @import("utils.zig");
const walkers = @import("walker.zig");
const BitWalkerUint = walkers.BitWalkerUint;
const BitWalker = walkers.BitWalker;

pub fn HuffmanGraph(comptime valueType: type) type {
    return struct {
        const Node = struct {
            const NodeSelf = @This();

            left: ?*NodeSelf,
            right: ?*NodeSelf,
            value: ?valueType,

            allocator: mem.Allocator,

            fn init(allocator: mem.Allocator) !*NodeSelf {
                var node = try allocator.create(NodeSelf);
                node.left = null;
                node.right = null;
                node.value = null;
                node.allocator = allocator;

                return node;
            }

            fn deinit(self: *NodeSelf) void {
                if (self.left) |left| {
                    left.deinit();
                }
                if (self.right) |right| {
                    right.deinit();
                }
                self.allocator.destroy(self);
            }

            fn print(self: *NodeSelf, curDepth: usize, targetDepth: usize) void {
                if (curDepth != targetDepth) {
                    if (self.left) |l| {
                        l.print(curDepth + 1, targetDepth);
                    } else {
                        utils.printf(" , ", .{});
                    }
                    if (self.right) |r| {
                        r.print(curDepth + 1, targetDepth);
                    } else {
                        utils.printf(" . ", .{});
                    }
                    return;
                }

                if (self.value) |v| {
                    utils.printf(" {any} ", .{v});
                } else {
                    utils.printf(" _ ", .{});
                }
            }

            fn depth(self: *NodeSelf) usize {
                var d: usize = 0;
                if (self.right) |r| {
                    d = r.depth();
                }
                if (self.left) |l| {
                    d = std.math.max(d, l.depth());
                }
                return d + 1;
            }

            fn valid(self: *NodeSelf) bool {
                var has_child = self.right == null and self.left == null;
                if (self.value != null)
                    return has_child;

                if (has_child)
                    return false;

                if (self.right) |r|
                    if (!r.valid())
                        return false;

                if (self.left) |l|
                    if (!l.valid())
                        return false;

                return true;
            }
        };

        root: *Node,
        allocator: mem.Allocator,
        stored_depth: ?usize = null,

        iterNode: *Node,

        const Self = @This();

        fn init(allocator: mem.Allocator) !Self {
            var root = try Node.init(allocator);
            return Self{
                .root = root,
                .iterNode = root,
                .allocator = allocator,
            };
        }

        pub fn print(self: *Self) void {
            var d = self.depth() + 1;
            for (0..d) |i| {
                self.root.print(0, i);
                utils.printf("\n", .{});
            }
        }

        fn addValue(self: *Self, code: u64, size: usize, value: valueType) !void {
            var walker = try BitWalkerUint(u64).init(code, size, true);

            var curNode: *Node = self.root;

            while (walker.walkBit()) |bit| {
                if (bit == 1) {
                    if (curNode.left) |nextNode| {
                        curNode = nextNode;
                    } else {
                        curNode.left = try Node.init(self.allocator);
                        curNode = curNode.left.?;
                    }
                } else {
                    if (curNode.right) |nextNode| {
                        curNode = nextNode;
                    } else {
                        curNode.right = try Node.init(self.allocator);
                        curNode = curNode.right.?;
                    }
                }
            }

            curNode.value = value;
            self.stored_depth = null;
        }

        fn depth(self: *Self) usize {
            if (self.stored_depth) |d| {
                return d;
            }

            self.stored_depth = self.root.depth() - 1;
            return self.stored_depth.?;
        }

        fn iter(self: *Self, bit: u1) anyerror!?valueType {
            var node = self.iterNode;

            var nextNode: ?*Node = null;

            if (bit == 1) {
                nextNode = node.left;
            } else {
                nextNode = node.right;
            }

            if (nextNode) |new_node| {
                if (new_node.value) |value| {
                    self.iterNode = self.root;
                    return value;
                } else {
                    self.iterNode = new_node;
                    return null;
                }
            }

            return error.InvalidBitSequence;
        }

        pub fn nextBitW(self: *Self, bitw: *BitWalker) !valueType {
            while (true) {
                if (try self.iter(try bitw.bitWalk())) |value|
                    return value;
            }
        }

        fn valid(self: *Self) bool {
            return self.root.valid();
        }

        pub fn deinit(self: *Self) void {
            self.root.deinit();
        }
    };
}

pub fn create_dynamic_graph(walker: *BitWalker, len_to_read: usize, allocator: mem.Allocator) !HuffmanGraph(u64) {
    const list: [19]u8 = .{ 16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15 };

    const lens: []u8 = try allocator.alloc(u8, 19);
    defer allocator.free(lens);
    mem.set(u8, lens, 0);

    for (0..len_to_read) |i| {
        lens[list[i]] = @intCast(u8, try walker.walk(3));
    }

    var builder = try GraphBuilder(u64).init(allocator, 19, 0, 8);
    defer builder.deinit();

    while (!builder.done()) {
        builder.addValue(builder.i, lens[builder.i]);
    }

    return try builder.build();
}

pub fn GraphBuilder(comptime T: type) type {
    return struct {
        const Self = @This();

        allocator: mem.Allocator,
        values: []T,
        lens: []usize,
        bl_count: []usize,
        size: usize,
        max: usize = 0,
        i: usize = 0,

        pub fn init(allocator: mem.Allocator, size: usize, default: T, maxlen: usize) !Self {
            var values = try allocator.alloc(T, size);
            mem.set(T, values, default);

            var lens = try allocator.alloc(usize, size);
            mem.set(usize, lens, 0);

            var bl_count = try allocator.alloc(usize, maxlen);
            mem.set(usize, bl_count, 0);

            return Self{
                .allocator = allocator,
                .bl_count = bl_count,
                .size = size,
                .values = values,
                .lens = lens,
            };
        }

        pub fn skip(self: *Self, len: usize) void {
            self.i += len;
        }

        pub fn done(self: *Self) bool {
            return self.i >= self.size;
        }

        pub fn addValue(self: *Self, value: T, len: usize) void {
            if (len > self.max) {
                self.max = len;
            }
            self.values[self.i] = value;
            self.lens[self.i] = len;
            self.bl_count[len] += 1;
            self.i += 1;
        }

        pub fn build(self: *Self) !HuffmanGraph(T) {
            self.bl_count[0] = 0;

            var next_code = try self.allocator.alloc(usize, self.max + 1);
            defer self.allocator.free(next_code);
            std.mem.set(usize, next_code, 0);

            var code: usize = 0;
            for (1..(self.max + 1)) |bits| {
                code = @shlExact(code + self.bl_count[bits - 1], 1);
                next_code[bits] = code;
            }

            var codes = try self.allocator.alloc(u64, self.size);
            defer self.allocator.free(codes);
            std.mem.set(usize, codes, 0);

            for (0..self.size) |n| {
                var len = self.lens[n];
                if (len != 0) {
                    codes[n] = next_code[len];
                    next_code[len] += 1;
                }
            }

            var graph = try HuffmanGraph(T).init(self.allocator);
            errdefer graph.deinit();

            for (0..self.size) |i| {
                if (self.lens[i] == 0)
                    continue;
                try graph.addValue(codes[i], self.lens[i], self.values[i]);
            }

            if (!graph.valid())
                return error.graph_not_valid;

            return graph;
        }

        pub fn buildWithDynamic(self: *Self, bitw: *BitWalker, graph: *HuffmanGraph(u64)) !HuffmanGraph(T) {
            while (!self.done()) {
                var v = try graph.nextBitW(bitw);

                if (v == 16) {
                    const len = try bitw.walk(2) + 3;
                    const last = self.lens[self.i - 1];
                    for (0..len) |_| {
                        self.addValue(self.i, last);
                    }
                } else if (v == 17) {
                    self.skip(try bitw.walk(3) + 3);
                } else if (v == 18) {
                    self.skip(try bitw.walk(7) + 11);
                } else {
                    self.addValue(self.i, v);
                }
            }

            return try self.build();
        }

        pub fn deinit(self: *Self) void {
            self.allocator.free(self.values);
            self.allocator.free(self.lens);
            self.allocator.free(self.bl_count);
        }
    };
}
