const std = @import("std");
const utils = @import("utils.zig");

pub fn BitWalkerUint(comptime T: anytype) type {
    const typeInfo = @typeInfo(T);

    if (typeInfo != .Int) {
        @compileError("This needs to be a int");
    }

    if (typeInfo.Int.signedness == .signed) {
        @compileError("The integer needs to be unsigned");
    }

    return struct {
        value: T,
        mask: T,

        // TODO this is probably wrong
        in_byte_position: i16,
        size: usize,
        reverse: bool,

        const Self = @This();

        pub fn init(value: T, size: usize, reverse: bool) !Self {
            if (typeInfo.Int.bits < size)
                return error.invlaid_size;

            var start_value: usize = 0;
            if (reverse) {
                start_value = size - 1;
            }

            var mask: T = 1;
            var i: u8 = 0;
            while (i < start_value) : (i += 1) {
                mask = @shlExact(mask, 1);
            }

            return Self{
                .value = value,
                .in_byte_position = @intCast(i16, start_value),
                .size = size,
                .mask = mask,
                .reverse = reverse,
            };
        }

        pub fn walkBit(self: *Self) ?u1 {
            if (self.mask == 0)
                return null;

            var result = (self.value & self.mask) == self.mask;

            //utils.print("walkbit: {b:08} & {b:08} -> {b:08} ({})", .{ self.value, self.mask, self.value & self.mask, result });

            if (self.reverse) {
                self.in_byte_position -= 1;
                if (self.mask == 1) {
                    self.mask = 0;
                } else {
                    self.mask = @shrExact(self.mask, 1);
                }
            } else {
                self.in_byte_position += 1;
                if (self.in_byte_position >= self.size) {
                    self.mask = 0;
                } else {
                    self.mask = @shlExact(self.mask, 1);
                }
            }

            return @boolToInt(result);
        }
    };
}

pub const BitWalker = struct {
    const Self = @This();

    data: *[]u8,

    // True  == most Significant
    // False == least Significant
    direction: bool = false,
    unitWaker: BitWalkerUint(u8),
    position: usize,

    pub fn init(data: *[]u8) !Self {
        return Self{
            .data = data,
            .position = 0,
            .unitWaker = try BitWalkerUint(u8).init(data.ptr[0], 8, false),
        };
    }

    pub fn change_direction(self: *Self) !void {
        self.direction = !self.direction;
        if (self.unitWaker.mask == 0)
            self.position += 1;
        // TODO probs wrong when mask != 0
        self.unitWaker = try BitWalkerUint(u8).init(self.data.ptr[self.position], 8, self.direction);
    }

    pub fn smart_change_direction(self: *Self) !void {
        self.direction = !self.direction;
        if (self.unitWaker.mask == 0) {
            self.position += 1;
            self.unitWaker = try BitWalkerUint(u8).init(self.data.ptr[self.position], 8, self.direction);
            return;
        }
        if (!self.direction) {
            self.unitWaker = try BitWalkerUint(u8).init(self.data.ptr[self.position], 8, self.direction);
        } else {
            self.unitWaker = try BitWalkerUint(u8).init(self.data.ptr[self.position], 8, self.direction);
        }
    }

    pub fn nextByte(self: *Self) !void {
        self.position += 1;
        self.unitWaker = try BitWalkerUint(u8).init(self.data.ptr[self.position], 8, self.direction);
    }

    pub fn status(self: *Self) void {
        utils.print("p: {}, in: {} ({b:08})", .{ self.position, self.unitWaker.in_byte_position, self.unitWaker.mask });
    }

    pub fn walk(self: *Self, bits: usize) !u64 {
        var number: u64 = 0;

        //utils.print("walk: {}", .{bits});

        var i: usize = 1;
        while (i <= bits) : (i += 1) {
            //utils.print("p: {}, in_p: {}", .{ self.position, self.unitWaker.in_byte_position });
            if (self.unitWaker.walkBit()) |next_number| {
                if (self.direction) {
                    number = @shlExact(number, 1);
                    number += next_number;
                } else {
                    number += @shlExact(@as(u64, next_number), @intCast(u6, i - 1));
                }
            } else {
                try self.nextByte();
                i -= 1;
            }
        }

        //utils.print("result: {} ({b:08})", .{ number, number });

        return number;
    }

    pub fn bitWalk(self: *Self) !u1 {
        return @intCast(u1, try self.walk(1));
    }
};
